import sqlite3
import atexit
import logging
from tkinter import E

class SQLHelper():
    def __init__(self, ignore_history, download_history):
        self.ignore_history = ignore_history
        self.download_history = download_history
        self.conn = None
        self.curs = None
        atexit.register(self.exit_handler) # Close the sql connection when the program exits

    async def sql_initialize(self):
        self.conn = sqlite3.connect(self.download_history)
        self.curs = self.conn.cursor()
        create_table_query = """CREATE TABLE IF NOT EXISTS downloads (
                                    filename TEXT,
                                    size INTEGER NOT NULL,
                                    completed INTEGER NOT NULL,
                                    PRIMARY KEY (filename, size)
                                );"""
        self.curs.execute(create_table_query)
        self.conn.commit()

    async def sql_check_existing(self, filename, size):
        if self.ignore_history:
            return False
        self.curs.execute("""SELECT completed FROM downloads WHERE filename = '%s' and size = %d""" % (filename, size))
        sql_file_check = self.curs.fetchone()
        if sql_file_check:
            if sql_file_check[0] == 1:
                return True
        return False

    async def sql_insert_file(self, filename, size, completed):
        self.curs.execute("""INSERT OR IGNORE INTO downloads VALUES ('%s', %d, %d)""" % (filename, size, completed))
        self.conn.commit()

    async def sql_update_file(self, filename, size, completed):
        self.curs.execute("""INSERT OR REPLACE INTO downloads VALUES ('%s', %d, %d)""" % (filename, size, completed))
        self.conn.commit()

    def exit_handler(self):
        try:
            self.conn.commit()
            self.conn.close()
        except Exception as e:
            logging.debug(f"Failed to close sqlite database connection: {str(e)}")
        else:
            logging.debug("Successfully closed sqlite database connection")
        


import requests

from colorama import Fore
from yarl import URL

from ..base_functions import log
from ..data_classes import DomainItem


class CyberfileCrawler:
    def __init__(self, username, password):
        self.base_url = URL('https://cyberfile.is/api/v2')
        self.username = username
        self.password = password
        self.token, self.account_id = self.create_account()

    def create_account(self):
        if not self.username or not self.password:
            log("username or password for cyberfile not provided")
            return False, False
        try:
            data = {"username": self.username, "password": self.password}
            res = requests.get(str(self.base_url/'authorize'), params=data)
            res.raise_for_status()
            res_json = res.json()
            if res_json["_status"] == "success":
                return res_json["data"]["access_token"], res_json["data"]["account_id"]
            else:
                raise requests.exceptions.RequestException(f'status: {res_json["_status"]}')
        except requests.exceptions.RequestException as e:
            log("Couldn't log into cyberfile")

    async def fetch(self, url: URL):
        domain_obj = DomainItem('cyberfile.is', {})
        if not self.username or not self.password or not self.account_id or not self.token:
            await log("Not logged into Cyberfile, skipping " + str(url))
            return domain_obj
        results = []
        await log("Starting scrape of " + str(url), Fore.WHITE)
        if 'folder' in url.parts:
            results.extend(await self.get_folder(url))
        else:
            await log("Skipping " + str(url) + " We can't currently handle single file links through the API")
        if results:
            for result in results:
                title, ret_url = result
                await domain_obj.add_to_album(title, ret_url, url)

        await log("Finished scrape of " + str(url), Fore.WHITE)
        return domain_obj

    async def get_folder(self, url: URL):
        try:
            data = {"access_token": self.token, "account_id": self.account_id, "parent_folder_id": url.parts[2]}
            res = requests.get(str(self.base_url / 'folder/listing'), params=data)
            res.raise_for_status()
            res_json = res.json()
            if res_json["_status"] == "success":
                for folder in res_json['data']['folders']:
                    if folder['folderName'] == url.parts[3]:
                        return await self.extract_folder(folder['id'], folder['folderName'])
            else:
                raise requests.exceptions.RequestException(f'status: {res_json["_status"]}')
        except requests.exceptions.RequestException as e:
            await log(e)

    async def extract_folder(self, id: int, title=None):
        results = []
        try:
            data = {"access_token": self.token, "account_id": self.account_id, "parent_folder_id": id}
            res = requests.get(str(self.base_url / 'folder/listing'), params=data)
            res.raise_for_status()
            res_json = res.json()
            if res_json["_status"] == "success":
                for folder in res_json['data']['folders']:
                    results.extend(await self.extract_folder(folder['id'], title+'/'+folder['folderName']))
                for file in res_json['data']['files']:
                    results.append((title, await self.download_link(file['id'])))
                return results
            else:
                raise requests.exceptions.RequestException(f'status: {res_json["_status"]}')
        except requests.exceptions.RequestException as e:
            await log(e)

    async def download_link(self, id: int):
        try:
            data = {"access_token": self.token, "account_id": self.account_id, "file_id": id}
            res = requests.get(str(self.base_url / 'file/download'), params=data)
            res.raise_for_status()
            res_json = res.json()
            if res_json["_status"] == "success":
                return URL(res_json['data']['download_url'])
            else:
                raise requests.exceptions.RequestException(f'status: {res_json["_status"]}')
        except requests.exceptions.RequestException as e:
            await log(e)


import asyncio
import logging
import multiprocessing
import sqlite3
import time
import traceback
from functools import wraps
from pathlib import Path
from typing import List, Optional, Tuple

import aiofiles
import aiofiles.os
import aiohttp.client_exceptions
from colorama import Fore
from tqdm import tqdm
from yarl import URL

from .base_functions import FILE_FORMATS, MAX_FILENAME_LENGTH, log, logger, sanitize, ssl_context, user_agent
from .sql_helper import SQLHelper
from .data_classes import AlbumItem, CascadeItem


class FailureException(Exception):
    """Basic failure exception I can throw to force a retry."""
    pass


def retry(f):
    @wraps(f)
    async def wrapper(self, *args, **kwargs):
        while True:
            try:
                return await f(self, *args, **kwargs)
            except FailureException:
                if not self.disable_attempt_limit:
                    if self.current_attempt >= self.attempts - 1:
                        raise
                logger.debug('Retrying %s...' % args[0])
                self.attempts += 1
                await asyncio.sleep(2)
    return wrapper


async def throttle(self, url: URL) -> None:
    host = url.host
    if host is None:
        return
    delay = self.delay.get(host)
    if delay is None:
        return

    key: Optional[str] = None
    while True:
        if key is None:
            key = 'throttle:{}'.format(host)
        now = time.time()
        last = self.throttle_times.get(key, 0.0)
        elapsed = now - last

        if elapsed >= delay:
            self.throttle_times[key] = now
            return

        remaining = delay - elapsed + 0.25

        log_string = '\nDelaying request to %s for %.2f seconds.' % (
            host, remaining)
        logger.debug(log_string)
        await asyncio.sleep(remaining)


class Downloader:
    def __init__(self, album_obj: AlbumItem, cookie_jar, folder: Path, title: str, attempts: int,
                 disable_attempt_limit: bool, max_workers: int, exclude_videos: bool, exclude_images: bool,
                 exclude_audio: bool, exclude_other: bool, SQL_helper: SQLHelper):
        self.album_obj = album_obj
        self.cookie_jar = cookie_jar
        self.folder = folder
        self.title = title

        self.SQL_helper = SQL_helper

        self.attempts = attempts
        self.current_attempt = 0
        self.disable_attempt_limit = disable_attempt_limit

        self.exclude_videos = exclude_videos
        self.exclude_images = exclude_images
        self.exclude_audio = exclude_audio
        self.exclude_other = exclude_other

        self.max_workers = max_workers
        self._semaphore = asyncio.Semaphore(max_workers)
        self.delay = {'media-files.bunkr.is': 0.5}
        self.throttle_times = {}

    """Changed from aiohttp exceptions caught to FailureException to allow for partial downloads."""

    @retry
    async def download_file(
            self,
            url: URL,
            referral: URL,
            filename: str,
            session: aiohttp.ClientSession,
            show_progress: bool = True
    ) -> None:
        """Download the content of given URL"""
        headers = {'Referer': str(referral), 'user-agent': user_agent}

        try:
            async with self._semaphore:
                if url.host in self.delay:
                    await throttle(self, url)

                ext = '.'+filename.split('.')[-1].lower()
                if not (ext in FILE_FORMATS['Images'] or ext in FILE_FORMATS['Videos'] or ext in FILE_FORMATS['Audio'] or ext in FILE_FORMATS['Other']):
                    async with session.get(url, headers=headers, ssl=ssl_context, raise_for_status=True) as resp:
                        filename = resp.content_disposition.filename
                        filename = await sanitize(filename)
                        del resp
                        if (self.folder / self.title / filename).exists():
                            return

                # Skip based on CLI arg.
                if self.exclude_videos:
                    if ext in FILE_FORMATS['Videos']:
                        logging.debug("Skipping " + filename)
                        return
                if self.exclude_images:
                    if ext in FILE_FORMATS['Images']:
                        logging.debug("Skipping " + filename)
                        return
                if self.exclude_audio:
                    if ext in FILE_FORMATS['Audio']:
                        logging.debug("Skipping " + filename)
                        return
                if self.exclude_other:
                    if ext in FILE_FORMATS['Other']:
                        logging.debug("Skipping " + filename)
                        return

                resume_point = 0
                complete_file = (self.folder / self.title / filename)
                temp_file = complete_file.with_suffix(
                    complete_file.suffix + '.part')

                if temp_file.exists():
                    resume_point = temp_file.stat().st_size
                    headers['Range'] = 'bytes=%d-' % resume_point

                async with session.get(url, headers=headers, ssl=ssl_context, raise_for_status=True) as resp:
                    content_type = resp.headers.get('Content-Type')
                    if 'text' in content_type.lower() or 'html' in content_type.lower():
                        logger.debug(
                            f"Server for %s is either down or the file no longer exists" % str(url))
                        return
                    total = int(resp.headers.get(
                        'Content-Length', str(0))) + resume_point

                    if await self.SQL_helper.sql_check_existing(filename, total):
                        logger.debug("%s Already Downloaded" % filename)
                        return

                    (self.folder / self.title).mkdir(parents=True, exist_ok=True)

                    await self.SQL_helper.sql_insert_file(filename, total, 0)

                    with tqdm(
                            total=total, unit_scale=True,
                            unit='B', leave=False, initial=resume_point,
                            desc=filename, disable=(not show_progress)
                    ) as progress:
                        async with aiofiles.open(temp_file, mode='ab') as f:
                            async for chunk, _ in resp.content.iter_chunks():
                                await f.write(chunk)
                                progress.update(len(chunk))
            resp.close()
            await self.rename_file(filename)
        except (aiohttp.client_exceptions.ClientPayloadError, aiohttp.client_exceptions.ClientOSError,
                aiohttp.client_exceptions.ServerDisconnectedError, asyncio.TimeoutError,
                aiohttp.client_exceptions.ClientResponseError, FailureException) as e:
            try:
                if 400 <= e.code < 500:
                    logger.debug(
                        "We ran into a 400 level error: %s" % str(e.code))
                    return
                resp.close()
            except:
                pass
            raise FailureException(e)

    async def rename_file(self, filename: str) -> None:
        """Rename complete file."""
        complete_file = (self.folder / self.title / filename)
        temp_file = complete_file.with_suffix(complete_file.suffix + '.part')
        if complete_file.exists():
            logger.debug(str(self.folder / self.title /
                         filename) + " Already Exists")
            await aiofiles.os.remove(temp_file)
        else:
            temp_file.rename(complete_file)
        await self.SQL_helper.sql_update_file(filename, complete_file.stat().st_size, 1)
        logger.debug("Finished " + filename)

    async def download_and_store(
            self,
            url_tuple: Tuple,
            session: aiohttp.ClientSession,
            show_progress: bool = True
    ) -> None:
        """Download the content of given URL and store it in a file."""
        url, referral = url_tuple

        filename = url.name
        filename = await sanitize(filename)
        if "v=" in filename:
            filename = filename.split('v=')[0]
        if len(filename) > MAX_FILENAME_LENGTH:
            fileext = filename.split('.')[-1]
            filename = filename[:MAX_FILENAME_LENGTH] + '.' + fileext

        complete_file = (self.folder / self.title / filename)
        if complete_file.exists():
            await self.SQL_helper.sql_update_file(filename, complete_file.stat().st_size, 1)
            logger.debug(str(complete_file) + " Already Exists")
        else:
            logger.debug("Working on " + str(url))
            try:
                await self.download_file(url, referral=referral, filename=filename,
                                         session=session, show_progress=show_progress)
            except Exception:
                logger.debug(traceback.format_exc())
                await log(f"\nError attempting {filename}: See downloader.log for details\n", Fore.WHITE)

    async def download_all(
            self,
            album_obj: AlbumItem,
            session: aiohttp.ClientSession,
            show_progress: bool = True
    ) -> None:
        """Download the data from all given links and store them into corresponding files."""
        coros = [self.download_and_store(url_object, session, show_progress)
                 for url_object in album_obj.link_pairs]
        for func in tqdm(asyncio.as_completed(coros), total=len(coros), desc=self.title, unit='FILES'):
            await func

    async def download_content(self, show_progress: bool = True) -> None:
        """Download the content of all links and save them as files."""
        async with aiohttp.ClientSession(cookie_jar=self.cookie_jar) as session:
            await self.download_all(self.album_obj, session, show_progress=show_progress)
        self.SQL_helper.conn.commit()

def get_downloaders(Cascade: CascadeItem, folder: Path, attempts: int, disable_attempt_limit: bool, threads: int,
                    exclude_videos: bool, exclude_images: bool, exclude_audio: bool, exclude_other: bool,
                    SQL_helper: SQLHelper) -> List[Downloader]:
    """Get a list of downloaders for each supported type of URLs.
    We shouldn't just assume that each URL will have the same netloc as
    the first one, so we need to classify them one by one, sort them to
    corresponding netloc URLs and create downloaders separately for individual
    netloc URLs they support.
    """

    downloaders = []

    cookie_jar = Cascade.cookies

    for domain, domain_obj in Cascade.domains.items():
        max_workers = threads if threads != 0 else multiprocessing.cpu_count()
        if 'bunkr' in domain:
            max_workers = 2 if (max_workers > 2) else max_workers
        for title, album_obj in domain_obj.albums.items():
            downloader = Downloader(album_obj, cookie_jar=cookie_jar, title=title, folder=folder,
                                    attempts=attempts, disable_attempt_limit=disable_attempt_limit,
                                    max_workers=max_workers, exclude_videos=exclude_videos,
                                    exclude_images=exclude_images, exclude_audio=exclude_audio,
                                    exclude_other=exclude_other, SQL_helper=SQL_helper)
            downloaders.append(downloader)
    return downloaders



import asyncio
import re

from bs4 import BeautifulSoup
from colorama import Fore
from yarl import URL

from ..base_functions import log, logger, make_title_safe, ssl_context
from ..data_classes import CascadeItem


class ThotsbayCrawler():
    def __init__(self, *, include_id=False, username=None, password=None, scraping_mapper, session, separate_posts=False):
        self.include_id = include_id
        self.separate_posts = separate_posts
        self.username = username
        self.password = password
        self.scraping_mapper = scraping_mapper
        self.session = session
        self.lock = 0

    async def login(self, session, url: URL):
        domain = URL("https://" + url.host) / "login"
        async with session.get(domain, ssl=ssl_context) as response:
            text = await response.text()
            if "You are already logged in" in text:
                return
            soup = BeautifulSoup(text, 'html.parser')

            inputs = soup.select('form input')
            data = {
                elem['name']: elem['value']
                for elem in inputs
                if elem.get('name') and elem.get('value')
            }
            data.update({"login": self.username, "password": self.password})
            return await session.post(domain/"login", data=data)

    async def fetch(self, session, url: URL):
        await log("Starting scrape of " + str(url), Fore.WHITE)
        Cascade = CascadeItem({})

        try:
            if self.username and self.password:
                while True:
                    if self.lock == 0:
                        self.lock = 1
                        await self.login(session, url)
                        self.lock = 0
                        break
                    else:
                        await asyncio.sleep(2)
            else:
                await log("login wasn't provided, consider using --thotsbay-username and --thotsbay-password")
                await log("Not being logged in might cause issues.")
        except Exception as e:
            await log("there was an error signing into %s" % url.host)
            await log(e)
            return

        try:
            title = await self.parse_thread(session, url, Cascade, None)
        except Exception:
            await log("Error handling " + str(url))
            return
        # await Cascade.append_title(title)

        await log("Finished scrape of " + str(url), Fore.WHITE)
        return Cascade

    async def parse_thread(self, session, url, Cascade, title):
        try:
            async with session.get(url, ssl=ssl_context) as response:
                text = await response.text()
                soup = BeautifulSoup(text, 'html.parser')

                domain = URL("https://" + url.host)

                title_block = soup.select_one("h1[class=p-title-value]")
                for elem in title_block.find_all("a"):
                    elem.decompose()

                if title:
                    pass
                else:
                    title = title_block.text
                    if self.include_id:
                        titlep2 = url.name
                        titlep2 = [s for s in titlep2 if "." in s][-1]
                        title = title + " - " + titlep2
                    title = await make_title_safe(title.replace(r"\n", "").strip())

                content_links = []

                post_number = str(url).split("post-")
                post_number = int(post_number[-1].strip("/")) if len(post_number) == 2 else None

                posts = soup.select("div[class='message-main uix_messageContent js-quickEditTarget']")
                for post in posts:
                    post_num_str = post.select_one("li[class=u-concealed] a").get('href').split('/')[-1]
                    post_num_int = int(post_num_str.split('post-')[-1])
                    if post_number:
                        if post_number > post_num_int:
                            continue

                    temp_title = title+"/"+post_num_str if self.separate_posts else title

                    for elem in post.find_all('blockquote'):
                        elem.decompose()
                    post_content = post.select_one("div[class=bbWrapper]")

                    # Content links
                    links = post_content.select('a')
                    for link in links:
                        link = link.get('href')
                        if link.endswith("/"):
                            link = link[:-1]
                        if link.startswith('//'):
                            link = "https:" + link
                        elif link.startswith('/'):
                            link = domain / link[1:]
                        content_links.append([URL(link), temp_title])

                    links = post.select("div[class='bbImageWrapper js-lbImage']")
                    for link in links:
                        link = link.get('data-src')
                        if link.endswith("/"):
                            link = link[:-1]
                        if link.startswith('/'):
                            link = domain / link[1:]
                        content_links.append([URL(link), temp_title])

                    links = post.select("div[class='bbImageWrapper lazyload js-lbImage']")
                    for link in links:
                        link = link.get('data-src')
                        if link.endswith("/"):
                            link = link[:-1]
                        if link.startswith('//'):
                            link = "https:" + link
                        elif link.startswith('/'):
                            link = domain / link[1:]
                        content_links.append([URL(link), temp_title])

                    links = post.select("video source")
                    for link in links:
                        link = link.get('src')
                        if link.endswith("/"):
                            link = link[:-1]
                        if link.startswith('//'):
                            link = "https:" + link
                        elif link.startswith('/'):
                            link = domain / link[1:]
                        content_links.append([URL(link), temp_title])

                    attachments_block = post.select_one("section[class=message-attachments]")
                    links = attachments_block.select("a[class='file-preview js-lbImage']") if attachments_block else []
                    for link in links:
                        link = link.get('href')
                        if link.endswith("/"):
                            link = link[:-1]
                        if link.startswith('/'):
                            link = domain / link[1:]
                        in_prog_title = temp_title + "/Attachments" if self.separate_posts else "Attachments"
                        await Cascade.add_to_album(url.host, in_prog_title, URL(link), url)
                    
                    # Gyfcat / Redgifs / Imgur
                    # NOTE IMGUR SUPPORT IS NOT WORKING AS OF NOW
                    links = post_content.select(
                        "span[data-s9e-mediaembed-iframe]")
                    for link in links:
                        embed_data = link.get("data-s9e-mediaembed-iframe")
                        # Replace gyfcat/redgifs/imgur embed url escape chars
                        embed_data = embed_data.replace("\/\/", "https://www.")
                        embed_data = embed_data.replace("\\", "")

                        # extract gyfcat/redgifs/imgur url from embed data
                        embed_url = re.search(
                            "https?:\/\/(www\.)?[-a-zA-Z0-9@:%._\+~#=]{2,256}\.[a-z]{2,4}\\b([-a-zA-Z0-9@:%_\+.~#?&//=]*)", embed_data)
                        if embed_url:
                            embed_url = URL(embed_url.group(
                                0).replace("www.", ""))
                            content_links.append([embed_url, temp_title])

                forum_direct_urls = [x for x in content_links if url.host in x[0].host]
                content_links = [x for x in content_links if x not in forum_direct_urls]
                for link_title_bundle in forum_direct_urls:
                    link = link_title_bundle[0]
                    temp_title = link_title_bundle[1]
                    in_prog_title = temp_title + "/Attachments" if self.separate_posts else "Attachments"
                    if str(link).endswith("/"):
                        link = URL(str(link)[:-1])
                    if 'attachments' in link.parts:
                        await Cascade.add_to_album(url.host, in_prog_title, link, url)
                    elif 'data' in link.parts:
                        await Cascade.add_to_album(url.host, in_prog_title, link, url)

                tasks = []
                for link_title_bundle in content_links:
                    link = link_title_bundle[0]
                    temp_title = link_title_bundle[1]
                    tasks.append(self.scraping_mapper.map_url(link, temp_title))
                await asyncio.gather(*tasks)

                next_page = soup.select_one('a[class="pageNav-jump pageNav-jump--next"]')
                if next_page is not None:
                    next_page = next_page.get('href')
                    if next_page is not None:
                        if next_page.startswith('/'):
                            next_page = domain / next_page[1:]
                        next_page = URL(next_page)
                        title = await self.parse_thread(session, next_page, Cascade, title)
                return title

        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            logger.debug(e)


import argparse
import asyncio
import logging
from pathlib import Path

from colorama import Fore
from yarl import URL

from . import __version__ as VERSION
from .utils.base_functions import clear, log, logger, purge_dir, regex_links
from .utils.downloaders import get_downloaders
from .utils.scraper import scrape
from .utils.sql_helper import SQLHelper


def parse_args():
    parser = argparse.ArgumentParser(description="Bulk downloader for multiple file hosts")
    parser.add_argument("-V", "--version", action="version", version="%(prog)s " + VERSION)
    parser.add_argument("-i", "--input-file", type=Path, help="file containing links to download", default="URLs.txt")
    parser.add_argument("-o", "--output-folder", type=Path, help="folder to download files to", default="Downloads")
    parser.add_argument("--log-file", help="log file to write to", default="downloader.log")
    parser.add_argument("--db-file", help="history database file to write to", default="download_history_size_based.sqlite")
    parser.add_argument("--threads", type=int, help="number of threads to use (0 = max)", default=0)
    parser.add_argument("--attempts", type=int, help="number of attempts to download each file", default=10)
    parser.add_argument("--disable-attempt-limit", help="disables the attempt limitation", action="store_true")
    parser.add_argument("--include-id", help="include the ID in the download folder name", action="store_true")
    parser.add_argument("--exclude-videos", help="skip downloading of video files", action="store_true")
    parser.add_argument("--exclude-images", help="skip downloading of image files", action="store_true")
    parser.add_argument("--exclude-audio", help="skip downloading of audio files", action="store_true")
    parser.add_argument("--exclude-other", help="skip downloading of images", action="store_true")
    parser.add_argument("--ignore-history", help="This ignores previous download history", action="store_true")
    parser.add_argument("--separate-posts", help="separates thotsbay scraping into folders by post", action="store_true")
    parser.add_argument("--thotsbay-username", type=str, help="username to login to thotsbay", default=None)
    parser.add_argument("--thotsbay-password", type=str, help="password to login to thotsbay", default=None)
    parser.add_argument("--cyberfile-username", type=str, help="username to login to cyberfile", default=None)
    parser.add_argument("--cyberfile-password", type=str, help="password to login to cyberfile", default=None)
    parser.add_argument("links", metavar="link", nargs="*", help="link to content to download (passing multiple links is supported)", default=[])
    args = parser.parse_args()
    return args


async def download_all(args: argparse.Namespace):
    await clear()
    await log(f"We are running version {VERSION} of Cyberdrop Downloader", Fore.WHITE)
    logging.debug(f"Starting downloader with args: {args.__dict__}")
    input_file = args.input_file
    if not input_file.is_file():
        input_file.touch()
        await log(f"{input_file} created. Populate it and retry.")
        exit(1)

    SQL_helper = SQLHelper(args.ignore_history, args.db_file)
    await SQL_helper.sql_initialize()

    links = args.links
    links = list(map(URL, links))

    with open(input_file, "r") as f:
        links += await regex_links(f.read())
    content_object = await scrape(links, args.include_id, args.thotsbay_username, args.thotsbay_password, args.cyberfile_username, args.cyberfile_password, args.separate_posts)
    if await content_object.is_empty():
        logging.error(f'ValueError No links')
        await log("No links found, check the URL.txt\nIf the link works in your web browser, please open an issue ticket with me.", Fore.RED)
        await log("This program does not currently support password protected albums.", Fore.RED)
        exit(0)
    await clear()
    downloaders = get_downloaders(content_object, folder=args.output_folder, attempts=args.attempts,
                                  disable_attempt_limit=args.disable_attempt_limit,
                                  threads=args.threads, exclude_videos=args.exclude_videos,
                                  exclude_images=args.exclude_images, exclude_audio=args.exclude_audio,
                                  exclude_other=args.exclude_other, SQL_helper=SQL_helper)

    for downloader in downloaders:
        await downloader.download_content()
    logger.debug("Finished")

    all_files = [str(f) for f in args.output_folder.glob("**/*") if f.is_file()]
    combined = '\t'.join(all_files)

    await log('Purging empty directories')
    deleted = await purge_dir(args.output_folder)

    await log('Finished downloading. Enjoy :)')
    if '.part' in combined:
        await log('There are still partial downloads in your folders, please re-run the program.')


def main(args = None):
    if args == None:
        args = parse_args()
    logging.basicConfig(
        filename=args.log_file,
        level=logging.DEBUG,
        format="%(asctime)s:%(levelname)s:%(module)s:%(filename)s:%(lineno)d:%(message)s",
        filemode="w"
    )
    asyncio.get_event_loop().run_until_complete(download_all(args))


if __name__ == '__main__':
    print("""
    STOP! If you're just trying to download files, check the README.md file for instructions.
    If you're developing this project, use start.py instead.
    """)
    exit()